/*
 *
 *
 *   risprinter.cc
 *
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


#include <iostream>
#include <fstream>
#include <string>

#include <cctype>
#include <curses.h>

#include "printer.h"


// allow empty lines
const static bool empty = true;

// work in std:: namespace
using namespace std;

// printer code constants (Epson fanfold line printer)
const std::string ESC = "\x1B";
const std::string PRINTER_ALARM = ESC + "x1" + ESC + "k1" + ESC + 'G' + ESC + '4';
const std::string PRINTER_NORMAL = ESC + "x0" + ESC + 'H' + ESC + '5';


/*
 *
 *   class RISPrinter
 *
 */

class RISPrinter : public Printer {
	private:  // options
		std::string alarm, normal;
		bool errors;

	private:  // state information
		bool escape;
		bool cr;

	public:   // public interface
		RISPrinter(const std::string &dev);
		void filter_print (char ch);

		void setAlarm(const std::string &s = "") { alarm = s; };
		void setNormal(const std::string &s = "") { normal = s; };

		void allowErrors(bool v = true) { errors = v; };

};


/*
 *
 *  RISPrinter::RISPrinter(...)
 *
 */

RISPrinter::RISPrinter(const std::string &dev)
	: Printer(dev) {
	errors = false;
	escape = false;
	cr = false;
}


/*
 *
 *   RISPrinter::filter_print(...)
 *
 */

void RISPrinter::filter_print (char ch) {
	static std::string out;

	if (ch == 0) return;

	switch (ch) {
		//case '\r':
		case '\n':
			if (escape)
				break;
			if (!cr) {
				out += "\r\n";
				try {
				print(out);
				} catch (...) {
					if (errors)
						throw;
				}
				out.erase();
			}
			cr = true;
			escape = false;
			break;

		case '\x1B':
			escape = true;
			cr = false;
			break;

		default:
			if (escape) {
				if ((ch == '3') && (! alarm.empty())) {
					out += alarm;
				} else if ((ch == '4') && (! normal.empty())) {
					out += normal;
				}
			} else {
				if (isprint(ch))
					out += ch;
			}

			if (empty)
				cr = false;
			escape = false;
			break;
	}
}


/*
 *
 *
 *   main()
 *
 *
 */

int main (int argc, char **argv) {
	if ((argc < 2) || (strlen(argv[1]) == 0)) {
		std::cerr << "Must specify printer device." << std::endl;
		return 0;
	}


	try {
		RISPrinter printer(argv[1]);
		char ch;

		printer.setAlarm(PRINTER_ALARM);
		printer.setNormal(PRINTER_NORMAL);

		while (std::cin) {
			std::cin.get(ch);
			if (ch)
				printer.filter_print(ch);
		}

	} catch (const std::exception &e) {
		std::cerr << e.what() << std::endl;
	}

	return 0;
}


// EOF
