/*
 *
 *
 *   sgcd.cc
 *
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


#include <fstream>
#include <iostream>
#include <string>
#include <list>

#include <my++/cmdline.h>
#include <my++/unixsocket.h>
#include <my++/select.h>

#include <cstring>
#include <cstdio>

#include "util.h"
#include "filter.h"
#include "signals.h"
#include "iobuf.h"

#include <syslog.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

// size of C-type I/O buffer
#define CBUF_SIZE (128)

// use std names by default
using namespace std;

// maximum size of either queue
const size_t queue_limit = 1024;



/*
 *
 *
 *    run_client(unixsocket&)
 *
 *
 */

static bool run_client (my::unixsocket &sock, bool network, bool readonly, bool dofilter) {
	SGFilter filter_obj;

	my::fdset rs;
	my::fdset ws;

	std::string message;

	iobuffer tx_sock_buf;
	iobuffer rx_sock_buf;

	char in_buffer [CBUF_SIZE];
	ssize_t rtn = 0;
	size_t limit;

	if (network)
		rx_sock_buf.push("+OK Go ahead.\r\n");

	for (;;) {
		// setup the multiplex
		rs.zero();
		ws.zero();

		if (!rx_sock_buf.full())
			rs.set(sock.getfd());
		if (!tx_sock_buf.full())
			rs.set(STDIN_FILENO);
		if (!tx_sock_buf.empty())
			ws.set(sock.getfd());
		if (!rx_sock_buf.empty())
			ws.set(STDOUT_FILENO);

		// wait for something to happen
		rtn = my::select(&rs, &ws, NULL, NULL);

		// deal with select errors
		if (rtn < 0) {
			if (errno == EINTR) {
				// if any signals occurred, bail
				if (signal_count(SIGHUP))
					return 0;
				if (signal_count(SIGTERM))
					return 0;
				if (signal_count(SIGINT))
					return 0;
				continue;
			} else {
				syslog(LOG_ERR, "Error occurred during select(): %s", strerror(errno));
				return true;
			}
		}

		if (rtn == 0)
			continue;

		if (rs.isset(sock.getfd())) {
			try {

				//
				//  READ from SOCKET
				//

				rtn = sock.get(in_buffer, sizeof(in_buffer));

				// if EOF reached, bail
				if (rtn == 0) {
					syslog(LOG_INFO, "Socket to sgd server closed");
					return true;
				}

				if (dofilter) {
					// build the output string
					message = filter_obj.filter(in_buffer, rtn);

					// insert the data into the buffer
					rx_sock_buf.push(message.c_str());
				} else {
					rx_sock_buf.push(in_buffer, rtn);
				}

			} catch (const my::device::exception &e) {
				if (e.errnum() != EINTR) {
					syslog(LOG_ERR, "Error reading from socket: %s", e.what());
					return true;
				}
			} catch (const my::unixsocket::exception &e) {
				if (e.errnum() != EINTR) {
					syslog(LOG_ERR, "Error reading from socket: %s", e.what());
					return true;
				}
			}
		}

		if (rs.isset(STDIN_FILENO)) {

			//
			//  READ from STDIN
			//

			rtn = read(STDIN_FILENO, in_buffer, sizeof(in_buffer));
			//rtn = read(STDIN_FILENO, tx_sock_buf);

			if (rtn < 0) {
				if (errno != EINTR) {
					syslog(LOG_ERR, "Error while reading from stdin: %s", strerror(errno));
					return false;
				}
			} else if (rtn == 0) {// if EOF reached, bail
				syslog(LOG_INFO, "Client connection closed");
				return false;
			} else {
				tx_sock_buf.insert(tx_sock_buf.end(), in_buffer, in_buffer + rtn);
			}
		}

		if (ws.isset(sock.getfd())) {
			if (tx_sock_buf.size() == 0) {
				syslog(LOG_WARNING, "INTERNAL: Received transmit event on socket, but no data in queue");
			} else {
				try {

					//
					//  write to SOCKET
					//

					if (readonly) {
						// if we are readonly, just pretend to write data :)
						rtn = tx_sock_buf.size();
					} else {
						// else, write the data to the socket
						rtn = sock.put(tx_sock_buf);
					}

					if (rtn > 0)
						tx_sock_buf.erase(tx_sock_buf.begin(), tx_sock_buf.begin() + rtn);

				} catch (const my::device::exception &e) {
					if (e.errnum() != EINTR) {
						syslog(LOG_ERR, "Error writing to socket: %s", e.what());
						return true;
					}
				} catch (const my::unixsocket::exception &e) {
					if (e.errnum() != EINTR) {
						syslog(LOG_ERR, "Error writing to socket: %s", e.what());
						return true;
					}
				}
			}
		}

		if (ws.isset(STDOUT_FILENO)) {
			if (rx_sock_buf.size() == 0) {
				syslog(LOG_WARNING, "INTERNAL: Received transmit event from client, but no data in queue");
			} else {

				//
				//  WRITE to CLIENT
				//

				limit = rx_sock_buf.size();
				if (limit > sizeof(in_buffer))
					limit = sizeof(in_buffer);

				std::copy(rx_sock_buf.begin(), rx_sock_buf.begin() + limit, in_buffer);

				rtn = write(STDOUT_FILENO, in_buffer, rtn);

				if (rtn < 0) {
					if (errno != EINTR) {
						syslog(LOG_ERR, "Error while writing to stdout: %s", strerror(errno));
						return false;
					}
				} else if (rtn == 0) {
					/* noop */
				} else {
					rx_sock_buf.erase(rx_sock_buf.begin(), rx_sock_buf.begin() + rtn);
				}
			}
		}
	}
}


static void usage (std::ostream &out) {
	out << endl;
	out << "Usage: sgcd [-pnrf] <socket_path>" << endl;
	out << "\t-f   enable filter mode" << endl;
	out << "\t-r   enable read-only mode" << endl;
	out << "\t-p   enable persistent mode" << endl;
	out << "\t-n   enable network mode (produce SGNET messages)" << endl;
	out << endl;
}


/*
 *
 *   main (...)
 *
 */

int main (int argc, char **argv) {
	// open syslog
	openlog("sgcd", LOG_PID, LOG_USER);

	// specifies a persistent client
	bool repeat   = false;

	// specifies a network client
	bool network  = false;

	// specifies read-only mode
	bool readonly = false;

	// enable filters
	bool filter   = false;

	my::command_line::config cfg;

	cfg.add("-p", my::command_line::arg_none);
	cfg.add("-n", my::command_line::arg_none);
	cfg.add("-r", my::command_line::arg_none);
	cfg.add("-f", my::command_line::arg_none);

	my::command_line::parser args(cfg);
	my::command_line::parser::nonoption_container nonopts;

	try {
		args.parse(argc, argv);
		args.get_nonoptions(nonopts);
	} catch (const std::exception &e) {
		if (network)
			cout << "-ERR Configuration Error\r\n";

		cerr << "Error: " << e.what() << endl << endl;
		usage(cerr);
		return 1;
	}

	if (args.is_set("-p"))
		repeat = true;

	if (args.is_set("-n"))
		network = true;

	if (args.is_set("-r"))
		readonly = true;

	if (args.is_set("-f"))
		filter   = true;

	// make sure the path was specified
	if (nonopts.size() != 1) {
		if (network)
			cout << "-ERR Configuration Error\r\n";

		usage(cerr);
		return 1;
	}

	my::unixsocket sock;
	std::string remote_address(getsockpath(nonopts.front()));

	// watch for signals
	setup_signals();

	do {
		try {

			//
			//  connect to remote
			//

			sock.connect(remote_address);

			// report success
			syslog (LOG_INFO, "Connected to %s", remote_address.c_str());

		} catch (const std::exception &e) {
			if (network)
				cout << "-ERR Connection failed: " << e.what() << "\r\n";
			syslog (LOG_ERR, "Error opening socket to sgd: %s", e.what());
			return 0;
		}

		// run one iteration
		repeat = run_client(sock, network, readonly, filter) && repeat;

		// no more banners
		network = false;

		// close the socket
		sock.close();

	} while (repeat);

	// complain
	syslog(LOG_INFO, "Exit");

	// should never reach here
	return 0;
}


// EOF
